/*
 * File:   main.c
 * Author: riddle
 *
 * Created on July 9, 2024, 1:37 AM

 * This code glitches an H8/300, HD6301Y0 or HD6301X0 into running external
 * code while internal ROM is enabled.  

 * The PIC has a much lower minimum voltage than HD6301Y0/X0
 * so if power is turned off while dumping, the 6301Y0 is likely to
 * output a few bad bytes before the PIC shuts down.

 * connections:
 * PIC     6301Y0  H8/300
 * ---     ------  ------
 * A0      MP0     MD0
 * A1      MP1     MD1
 * A2      /reset  /RES
 * A3      /stby   /STBY
 * A4      /nmi    /NMI
 * A5      VCC     VCC

 * C2      EXTAL   EXTAL

 * D7      P6.7    P5.2
 * D6      P6.6    P6.6
 * D5      P6.5    P6.5
 * D4      P6.4    P6.4
 * D3      P6.3    P6.3
 * D2      P6.2    P6.2
 * D1      P6.1    P6.1
 * D0      P6.0    P6.0

 * B0      P2.0    P5.0

 * 6301Y0  H8/300  EPROM
 * ------  ------  -----
 * P3.0    P3.0    D0 via 100 ohm resistor
 * P3.1    P3.1    D1 via 100 ohm resistor
 * P3.2    P3.2    D2 via 100 ohm resistor
 * P3.3    P3.3    D3 via 100 ohm resistor
 * P3.4    P3.4    D4 via 100 ohm resistor
 * P3.5    P3.5    D5 via 100 ohm resistor
 * P3.6    P3.6    D6 via 100 ohm resistor
 * P3.7    P3.7    D7 via 100 ohm resistor

 * P1.0    P1.0    A0
 * P1.1    P1.1    A1
 * P1.2    P1.2    A2
 * P1.3    P1.3    A3
 * P1.4    P1.4    A4
 * P1.5    P1.5    A5
 * P1.6    P1.6    A6
 * P1.7    P1.7    A7

 * PIC     EPROM
 * ---     -----
 * B4      A8
 * B5      A9
 * B6      A10

 * EPROM /CS, /OE - tie low - if using > 2K EPROM, tie A11 and above low
 * 6301Y0 /HALT, IRQ1, IRQ2, MR - tie high
 * H8/300 /WAIT - tie high
 * no decoupling caps

 * 115,200 baud
 * PIC serial out - 1K resistor on C6 - no level shifter needed direct to RS-232 Rx
 * PIC serial in - 22K resistor on C7 - no level shifter needed direct to RS-232 Tx

 * If you want to use a TTL-to-USB adapter, connect B.1 to ground with a 1K resistor; then PIC C6 is
 * TTL out and PIC C7 is TTL in.

 * 6301X0 dumps 16K but only last 4K are internal ROM.
 * at 115,200 bps, 16K dump takes about 8 seconds.

 * H8/300 dumps 64K, internal ROM space varies by chip.
 * H8/300 dump may not start at $0000, so dump entire address space twice (about 2 minutes),
 * then locate start by looking for external EPROM contents in external address space locations.

 * For example, H8/325 has 32K of internal ROM, then 31,616 bytes are external, then 1K of internal RAM,
 * then 16 bytes external, then 16 bytes of registers, then 16 bytes external, then 80 bytes of registers.
 * The H8/300 dumping code in the external ROM will be dumped whenever the external address space is selected.

 * The external EPROM is partitioned into eight 256-byte areas.
 * Currently only 2 of the areas are used; the first (A8=0, A9=0, A10=0) is selected for HD6301X0/HD6301Y0 dumping.
 * The second (A8=1, A9=0, A10=0) is selected for H8/300 dumping.  The other areas will be used as needed for
 * other chips.
 */

#define _XTAL_FREQ 64000000U
#include <xc.h>
#include <stdio.h>
#include "mcc_generated_files/system/pins.h"
#include "mcc_generated_files/system/system.h"
#include "serial.h"

#define MODEPOS 0
#define NRESETPOS 2
#define NSTBYPOS 3
#define NNMIPOS 4
#define POWERPOS 5

bool dodebug=false;
char debugstr[255];

uint8_t curlata=0;
void setsignals(bool curpower,bool curnnmi, bool curnstby, bool curnreset, uint8_t curmode) {
    curlata=(uint8_t)((curpower<<POWERPOS)|(curnnmi<<NNMIPOS)|(curnstby<<NSTBYPOS)|(curnreset<<NRESETPOS)|(curmode<<MODEPOS));
    LATA=curlata;
    if(dodebug) {
        sprintf(debugstr,"ss %x",curlata);
        sputs(debugstr);
    }
}

void setnreset(bool newnreset) {
    curlata&=(~(1<<NRESETPOS));
    curlata|=(newnreset<<NRESETPOS);
    LATA=curlata;
    if(dodebug) {
        sprintf(debugstr,"snr %x",curlata);
        sputs(debugstr);
    }
}

void setpower(bool newpower) {
    curlata&=(~(1<<POWERPOS));
    curlata|=(newpower<<POWERPOS);
    LATA=curlata;
    if(dodebug) {
        sprintf(debugstr,"sp %x",curlata);
        sputs(debugstr);
    }
}

void setnstby(bool newnstby) {
    curlata&=(~(1<<NSTBYPOS));
    curlata|=(newnstby<<NSTBYPOS);
    LATA=curlata;
    if(dodebug) {
        sprintf(debugstr,"sns %x",curlata);
        sputs(debugstr);
    }
}

void setnnmi(bool newnnmi) {
    curlata&=(~(1<<NNMIPOS));
    curlata|=(newnnmi<<NNMIPOS);
    LATA=curlata;
    if(dodebug) {
        sprintf(debugstr,"snn %x",curlata);
        sputs(debugstr);
    }
}

void setmode(uint8_t newmode) {
    curlata&=(~(3<<MODEPOS));
    curlata|=(newmode<<MODEPOS);
    LATA=curlata;
    if(dodebug) {
        sprintf(debugstr,"sm %x",curlata);
        sputs(debugstr);
    }
}

uint8_t eprombank=0;
void seteprombank(int b) {
    eprombank=b&7;
    eproma8_PORT=eprombank&1;
    eproma9_PORT=(eprombank&2)>>1;
    eproma10_PORT=(eprombank&4)>>2;
    if(dodebug) {
        sprintf(debugstr,"eb %d",eprombank);
        sputs(debugstr);
    }
}

int curclk=0;
void setclk(int speed) {
    if(speed==0) {        
        T2CON=0;
    } else if(speed<0) {
        T2CON=4;
    } else {
        TRISC|=4;
        CCPTMRS0=0;
        switch(speed) {
            case 8000:
                PR2=0x01;
                CCP1CON=0x3C;
                CCPR1=0x00;
                break;
            case 4000:
                PR2=0x03;
                CCP1CON=0x3C;
                CCPR1=0x01;
                break;
            case 2000:
                PR2=0x07;
                CCP1CON=0x3C;
                CCPR1=0x03;
                break;
            case 1000:
                PR2=0x0f;
                CCP1CON=0x3C;
                CCPR1=0x07;
                break;
            case 500:
                PR2=0x1f;
                CCP1CON=0x3C;
                CCPR1=0x0f;
                break;
            case 250:
                PR2=0x3f;
                CCP1CON=0x3C;
                CCPR1=0x1f;
                break;
        }
        T2CON=4;
        TRISC&=0xfb;
    }
    curclk=speed;
    if(dodebug) {
        sprintf(debugstr,"sc %d",curclk);
        sputs(debugstr);
    }
}

uint8_t linelen=0;
const char* hexDigits = "0123456789ABCDEF";
uint8_t portdval;

//this interrupt routine reads the data from the H8 or HD6301Y0 on
//the rising edge of BO and outputs it via the PIC's serial TX pin
//serial stuff is in-lined to speed up ISR
void int0handler() {   
    portdval=PORTD;
    while(!(PIR1bits.TX1IF && TXSTA1bits.TXEN)) {        
    }
    TXREG1=hexDigits[(portdval & 0xf0) >> 4];
    
    //TXxIF not valid until 2nd instruction following write
    asm("nop");
    
    while(!(PIR1bits.TX1IF && TXSTA1bits.TXEN)) {        
    }
    TXREG1=hexDigits[portdval & 0x0f];

    //need to do this anyway, so move here instead of NOP
    linelen+=16;    //256/16=16 hex digits per line

    while(!(PIR1bits.TX1IF && TXSTA1bits.TXEN)) {        
    }
    TXREG1=' ';

    if(linelen==0) {
        while(!(PIR1bits.TX1IF && TXSTA1bits.TXEN)) {        
        }
        TXREG1='\n';
    }
}

void main(void) {
    char sercmd;
    bool showver;
    char dbcmd;
    bool showdb;
    uint8_t oldlinelen;
    uint16_t lcount;
    uint8_t dcount;
    uint8_t retrying;

    SYSTEM_Initialize();
    RBPU=0;                          //weak pullup on B1 for use232adapter
    PLLEN = 1;                       //1=4x PLL = 64MHz
    while(PLLRDY==0) ;               //wait until system clock using PLL
    
    Init_Serial(use232adapter_PORT);

    __delay_ms(1000);
    
    EXT_INT0_InterruptDisable();
    INT0_SetInterruptHandler(int0handler);
    INTERRUPT_GlobalInterruptEnable();
    INTERRUPT_PeripheralInterruptEnable();

    seteprombank(0);
    setclk(0);

top:
    curlata=LATA;
    linelen=0;
    showver=1;
    showdb=1;
    retrying=0;
    dcount=0;
    
    while(1) {
        bool nokey=true;
        while(nokey) {
            //mloop:
            if(showver==1) {
                sputs("");
                sputs("dumph8and6301y0 2.0");
                sputs("1: dump HD6301Y0/HD6301X0");
                sputs("2: dump HD6301X0 method 2");
                sputs("3: dump H8/3214, H8/3292");
                sputs("4: dump H8/323");
                sputs("5: dump H8/3258");
                sputs("6: dump H8/3258 #2");
                sputs("7: dump H8/3257, H8/3256");
                sputs("8: dump H8/3044");
                sputs("9: dump H8/3258 #3a");
                sputs("D: debug mode");
                sputs("?: this message");
                showver=0;
            }
            if(EUSART1_IsRxReady()) {
                sercmd=sgetch();
                if(sercmd>='a') {
                    if(sercmd<='z') {
                        sercmd-=32;
                    }
                }
                if(sercmd=='?') {
                    showver=1;
                } else {
                    nokey=false;
                }
            }
        }//while nokey
        
retry:
        EXT_INT0_InterruptDisable();
        //menu items here
        switch(sercmd) {
            case '1':   //Y0 and X0 method 1
                sputs("6301Y0 dump");

                //select 6301Y0/X0 dumping code    
                seteprombank(0);

                //mode 0, reset active, power on
                setsignals(1,1,1,0,0);//power,nnmi,nstby,nreset,mode
                setclk(250);
                __delay_ms(100);
                EXT_INT0_InterruptEnable();

                //reset inactive, mode 2
                setnreset(1);
                setmode(2);

                __delay_us(65);
                
                //turn clock off
                setclk(0);

                __delay_us(2);

                //reset active
                setnreset(0);

                //reset inactive
                setnreset(1);

                __delay_us(65);

                //turn clock back on
                setclk(-1);
                break;
                
            case '2':
                sputs("6301X0 dump");
                //select 6301Y0/X0 dumping code    
                seteprombank(0);

                //mode 0, reset active, power on
                setsignals(1,1,1,0,0);//power,nnmi,nstby,nreset,mode
                setclk(250);
                __delay_ms(1000);
                setnreset(1);
                setsignals(1,1,1,1,2);//power,nnmi,nstby,nreset,mode
                setclk(0);//stop clock
                __delay_us(2);
                setnreset(0);
                __delay_ms(200);
                setnreset(1);
                setclk(-1); //resume clock
                setpower(0);
                __delay_ms(500);
                setpower(1);
                __delay_ms(500);
                
                EXT_INT0_InterruptEnable();
                
                setsignals(1,1,1,1,3);//power,nnmi,nstby,nreset,mode
                __delay_ms(500);
                setsignals(1,1,1,1,0);//power,nnmi,nstby,nreset,mode
                __delay_ms(500);
                setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode
                __delay_ms(1000);
                setpower(1);
                break;
                
            case '3':
                sputs("H8/3214, H8/3292 dump");

                //select H8/300 dumping code, $8080 reset vector
                seteprombank(1);

                //power on, mode 1, unreset, unstandby, unnmi        
                setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode
                setclk(250);
                __delay_ms(1000);
                EXT_INT0_InterruptEnable();
                setmode(2);
                break;
                
            case '4':
                sputs("H8/323 dump");

                //select H8/300 dumping code, $8080 reset vector
                eproma8_PORT=1;
                eproma9_PORT=0;
                eproma10_PORT=0;

                //power on, mode 1, unreset, unstandby, unnmi        
                setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode
                setclk(250);
                __delay_ms(1000);
                EXT_INT0_InterruptEnable();
                setmode(2);
                setnreset(0);
                __delay_ms(100);
                setnreset(1);
                __delay_ms(100);
                setpower(0);
                __delay_ms(100);
                setpower(1);
                break;
                
            case '5':
                sputs("H8/3258 dump");

                if(retrying) {
                    //PWM off
                    setclk(0);
                    __delay_ms(200);
    
                    setnreset(0);
                    setpower(0);
                    __delay_ms(1000);
                }

                //select H8/300 dumping code, $8080 reset vector
                seteprombank(1);

                //power on, mode 1, unreset, unstandby, unnmi
                setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode

                //PWM 4MHz 50% duty cycle from C.2 to EXTAL
                setclk(4000);
                __delay_ms(1000);

                setmode(2);

                __delay_ms(1500);

                setnreset(0);
                __delay_ms(500);
                setnreset(1);
                __delay_ms(2000);
                setpower(0);
                __delay_ms(500);
                setpower(1);

                //slow down so enough time to transmit data serially    
                //PWM 250kHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                setclk(250);

                EXT_INT0_InterruptEnable();
                break;
                
            case '6':
                sputs("H8/3258 dump #2");

                if(retrying) {
                    //PWM off
                    setclk(0);
                    __delay_ms(200);
    
                    setnreset(0);
                    setpower(0);
                    __delay_ms(1000);
                }

                //select H8/300 dumping code, $8080 reset vector
                seteprombank(1);

                //power on, mode 1, unreset, unstandby, unnmi        
                setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode
                setclk(2000);

                __delay_ms(1000);

//                EXT_INT0_InterruptEnable();

                setmode(2);    
                __delay_ms(1500);

                setnreset(0);
                __delay_ms(170);
                setnreset(1);
                __delay_ms(550);                
                setnnmi(0);
                __delay_ms(170);
                setnnmi(1);
                __delay_ms(550);
                setpower(0);
                __delay_ms(170);
                setpower(1);

                //slow down so enough time to transmit data serially    
                setclk(250);
  
                EXT_INT0_InterruptEnable();
                break;
                
            case '7':
                sputs("H8/3257, H8/3256 dump");

                if(!retrying) {
                    //select H8/300 dumping code, $F0F0 reset vector
                    seteprombank(2);
        
                    //power on, mode 1, unreset, unstandby, unnmi        
                    setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode   
    
                    setclk(4000);
                    __delay_ms(1000);
//                    EXT_INT0_InterruptEnable();

                    setmode(2);
                    __delay_ms(500);
    
                    setclk(0);
                    __delay_ms(200);
    
                    //slow down so enough time to transmit data serially    
                    setclk(250);
                }

                setnreset(0);
                __delay_ms(100);
                setnreset(1);
                __delay_ms(100);

                setpower(0);
                __delay_ms(100);
                setpower(1);
                EXT_INT0_InterruptEnable();
                break;
                
            case '8':
                sputs("H8/3044 dump");
                //select H8/300H dumping code
                seteprombank(3);

                //power on, mode 3, unreset, unstandby, unnmi -chip md1 tied high, sig md1 to chip md2
                setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode   

                //PWM 4MHz 50% duty cycle from C.2 to EXTAL
                setclk(4000);
    
                __delay_ms(1000);
//    INTCON=$90                      ;enable INT0 (portb.0)

                //mode 6
                setmode(2);

                //slow down so enough time to transmit data serially    
                //PWM 250kHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                setclk(250);
                EXT_INT0_InterruptEnable();
                break;
    
            case '9':
                sputs("H8/3258 dump #3a");

                if(retrying) {
                    //PWM off
                    setclk(0);
                    __delay_ms(200);
    
                    setnreset(0);
                    setpower(0);
                    __delay_ms(1000);
                }

                //select H8/300 dumping code, $8080 reset vector
                seteprombank(1);

                //power on, mode 1, unreset, unstandby, unnmi
                setsignals(1,1,1,1,1);//power,nnmi,nstby,nreset,mode

                //PWM 2MHz 50% duty cycle from C.2 to EXTAL
                setclk(2000);
                __delay_ms(1000);
//    INTCON=$90                      ;enable INT0 (portb.0)

                setmode(2);
 
                __delay_ms(1500);

                setnreset(0);
                __delay_ms(450);
                setnreset(1);
                __delay_ms(2000);//max delay < 4000ms
                __delay_ms(2000);
                setpower(0);
                __delay_ms(400);
                setnnmi(0);
                __delay_ms(450);
                setpower(1);
                __delay_ms(400);
                setnnmi(1);
    
                //slow down so enough time to transmit data serially    
                //PWM 250kHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                setclk(250);
                __delay_ms(200);
                EXT_INT0_InterruptEnable();
                break;
                                
            case 'D':
                nokey=true;
                while(nokey) {
                    if(showdb) {
                        sputs("");
                        sputs("debug mode");

                        sprintf(debugstr,"mode %d",curlata&3);
                        sputs(debugstr);

                        sprintf(debugstr,"bank %d",eprombank);
                        sputs(debugstr);

//                        if(power_PORT==1)
                        if(curlata&(1<<POWERPOS))
                            sputs("power on");
                        else
                            sputs("power off");
                        
//                        if(nreset_PORT==0)
                        if(curlata&(1<<NRESETPOS))
                            sputs("reset inactive");
                        else
                            sputs("reset active");
//                        if(nnmi_PORT==0)
                        if(curlata&(1<<NNMIPOS))
                            sputs("NMI inactive");
                        else
                            sputs("NMI active");
//                        if(nstby_PORT==0)
                        if(curlata&(1<<NSTBYPOS))
                            sputs("standby inactive");
                        else
                            sputs("standby active");

                        sprintf(debugstr,"osc %d",curclk);
                        sputs(debugstr);

                        sputs("");
                        sputs("M: change H8 mode");
                        sputs("E: change external EPROM bank");
                        sputs("P: toggle power");
                        sputs("R: toggle reset");
                        sputs("N: toggle NMI");
                        sputs("S: toggle stby");

                        sputs("0: stop osc");
                        sputs("1: 1 MHz");
                        sputs("2: 2 MHz");
                        sputs("4: 4 MHz");
                        sputs("8: 8 MHz");
                        sputs("9: 250 kHz");

                        sputs("D: dump to serial");
                        sputs("Q: reset all parameters");
                        sputs("X: exit debug mode");
                        sputs("?: this message");

                        showdb=0;
                    }
                    if(EUSART1_IsRxReady()) {
                        dbcmd=sgetch();
                        //uppercase command
                        if(dbcmd>='a'&&dbcmd<='z') {
                            dbcmd-=32;
                        }
                        if(dbcmd=='?') {
                            showdb=1;
                        } else {
                            nokey=false;
                        }
                    }
                }
                if(dbcmd=='X')   {//exit debug mode
                    goto top;
                }
                if(dbcmd=='D')   {//dump to screen
                    setclk(250);
                    EXT_INT0_InterruptEnable();
                    while(1) {}
                }
                if(dbcmd=='Q')   {//reset everything
                    showdb=1;
                    seteprombank(0);
                    setsignals(0,0,0,0,0);
                    setclk(0);
                }
                if(dbcmd=='M')   {
                    setmode(((curlata&3)+1)&3);
                    sprintf(debugstr,"mode %d",curlata&3);
                    sputs(debugstr);
                }
                if(dbcmd=='E')   {
                    seteprombank((eprombank+1)&7);
                    sprintf(debugstr,"bank %d",eprombank);
                    sputs(debugstr);
                }
                if(dbcmd=='R')   {
                    if(curlata&(1<<NRESETPOS)) {
//                    if(nreset_PORT==1) {
                        sputs("reset active");
                        setnreset(0);
                    } else {
                        sputs("reset inactive");
                        setnreset(1);
                    }
                }
                if(dbcmd=='S')   {
                    if(curlata&(1<<NSTBYPOS)) {
//                    if(nstby_PORT==1) {
                        sputs("standby active");
                        setnstby(0);
                    } else {
                        sputs("standby inactive");
                        setnstby(1);
                    }
                }
                if(dbcmd=='N')   {
                    if(curlata&(1<<NNMIPOS)) {
//                    if(nnmi_PORT==1) {
                        sputs("NMI active");
                        setnnmi(0);
                    } else {
                        sputs("NMI inactive");
                        setnnmi(1);
                    }
                }
                if(dbcmd=='P')   {
                    if(curlata&(1<<POWERPOS)) {
//                    if(power_PORT==1) {
                        sputs("power off");
                        setpower(0);
                    } else {
                        sputs("power on");
                        setpower(1);
                    }
                }
                if(dbcmd=='0')   {
                    //PWM off
                    setclk(0);
                    sputs("osc stopped");
                }
                if(dbcmd=='9')   {
                    //PWM 250kHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                    setclk(250);
                    sputs("250 kHz");
                }
                if(dbcmd=='1')   {
                    //PWM 1MHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                    setclk(1000);
                    sputs("1 MHz");
                }
                if(dbcmd=='2')   {
                    //PWM 2MHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                    setclk(2000);
                    sputs("2 MHz");
                }
                if(dbcmd=='4')   {
                    //PWM 4MHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                    setclk(4000);
                    sputs("4 MHz");
                }
                if(dbcmd=='8')   {
                    //PWM 8MHz 50% duty cycle from PIC C.2 to H8/300 EXTAL
                    setclk(8000);
                    sputs("8 MHz");
                }
    
                goto retry;
                break;
        }//switch

        //sit here while the ROM is dumped with INT0
        retrying=0;
        oldlinelen=linelen;
        lcount=0;
        while(true) {
            if(dcount<16) { //if we haven't received much data, then check for timeout
                if(linelen==oldlinelen) {
                    if(lcount==65535) {
                        retrying=1;
                        sputs(" retrying ");
                        goto retry;
                    }
                    lcount++;
                } else {
                    lcount=0;
                    if(dcount<16) {
                        dcount++;
                    }
                }
                oldlinelen=linelen;
            }
        }

    }//main while
    
    //eliminate some "function is never called" Warning 520s
    PIN_MANAGER_IOC();
    INT1_ISR();
    INT1_CallBack();
    INT2_ISR();
    INT2_CallBack();
}//main

